package ap;

import ap.proof.{ConstraintSimplifier, ModelSearchProver, Timeout}
import ap.proof.tree.ProofTree
import ap.terfor.conjunctions.{Conjunction, Quantifier}
import ap.parameters.{GoalSettings, PreprocessingSettings}

object IntelliFileProver {
  
  abstract class Result
  
  abstract class ProofResult extends Result
  case class Proof(tree : ProofTree, model : Conjunction) extends ProofResult
  case class NoProof(unsatisfiableTree : ProofTree) extends ProofResult
  case class TimeoutProof(unfinishedTree : ProofTree) extends ProofResult

  // "model" means that the implicitly existentially quantified constants can be
  // instantiated so that the formula evaluates to true
  abstract class ModelResult extends Result
  case class Model(model : Conjunction) extends ModelResult
  case object NoModel extends ModelResult
  case object TimeoutModel extends ModelResult

  // "countermodel" means that the implicitly universally quantified constants
  // can be instantiated so that the formula evaluates to false
  abstract class CounterModelResult extends Result
  case class CounterModel(counterModel : Conjunction) extends CounterModelResult
  case object NoCounterModel extends CounterModelResult
  case object TimeoutCounterModel extends CounterModelResult
  
}

/**
 * A prover that decides, depending on the kind of the problem, whether it
 * should try to construct a proof tree or just search for counterexamples
 */
class IntelliFileProver(reader : java.io.Reader,
                        // a timeout in milliseconds
                        timeout : Int,
                        mostGeneralConstraint : Boolean,
                        output : Boolean,
                        userDefStoppingCond : => Boolean,
                        preprocSettings : PreprocessingSettings,
                        goalSettings : GoalSettings)
      extends AbstractFileProver(reader, output, timeout,
                                 userDefStoppingCond, preprocSettings,
                                 goalSettings) {

  import IntelliFileProver._
        
  lazy val proofResult : ProofResult =
    Timeout.catchTimeout[ProofResult] {
      val tree = constructProofTree(mostGeneralConstraint)
      val model = findModel(tree.closingConstraint)
      if (model.isFalse)
        NoProof(tree)
      else
        Proof(tree, model)
    } {
      // we assume that a partial proof tree is given back
      case x : ProofTree => TimeoutProof(x)
    }
        
  lazy val proofTree : ProofTree = proofResult match {
    case TimeoutProof(t) => t
    case Proof(t, _) => t
    case NoProof(t) => t
  } 

  lazy val modelResult : ModelResult =
    Timeout.catchTimeout[ModelResult] { 
      val model = findModelTimeout
      if (model.isFalse)
        NoModel
      else
        Model(model)
    } {
      case _ => TimeoutModel
    }

  lazy val counterModelResult : CounterModelResult =
    Timeout.catchTimeout[CounterModelResult] { 
      val model = findCounterModelTimeout
      if (model.isFalse)
        NoCounterModel
      else
        CounterModel(model)
    } {
      case _ => TimeoutCounterModel
    }

  val result : IntelliFileProver.Result = {
    val quantifiers = Conjunction.collectQuantifiers(formula)
    if (!mostGeneralConstraint &&
        eliminatedConstants.isEmpty && formula.predicates.isEmpty &&
        (quantifiers subsetOf Set(Quantifier.EX))) {
      // try to find a model
      modelResult
    } else if ((formula.constants subsetOf eliminatedConstants) &&
               (quantifiers subsetOf Set(Quantifier.ALL))) {
      // try to find a countermodel
      counterModelResult
    } else {
      // try to construct a proof
      proofResult
    }
  }
  
}
