package ap.proof;

import ap.basetypes.IdealInt
import ap.terfor.{Formula, TermOrder, ConstantTerm}
import ap.terfor.conjunctions.{Conjunction, Quantifier}
import ap.terfor.linearcombination.LinearCombination
import ap.terfor.equations.EquationConj
import ap.terfor.substitutions.{Substitution, IdentitySubst}
import ap.proof.goal.{Goal, FactsNormalisationTask, EliminateFactsTask,
                      AddFactsTask, UpdateTasksTask}
import ap.util.{Debug, Logic, LRUCache, FilterIt, Seqs}
import ap.parameters.{GoalSettings, Param}
import ap.proof.tree._

import scala.collection.mutable.Stack


/**
 * A prover that tries to construct a countermodel of a ground formula. This
 * prover works in depth-first mode, in contrast to the
 * <code>ExhaustiveProver</code>.
 */
object ModelSearchProver {

  private val AC = Debug.AC_PROVER
   
  private val simplifier = ConstraintSimplifier.FAIR_SIMPLIFIER
  
  // we need to store eliminated facts from goals, otherwise we could not
  // construct a complete model
  private val ptf = new SimpleProofTreeFactory(true, simplifier) {
    override def eliminatedConstant(subtree : ProofTree,
                                    c : ConstantTerm,
                                    witness : (Substitution, TermOrder) => Substitution,
                                    vocabulary : Vocabulary) : ProofTree =
      new WitnessTree (subtree, c, witness, vocabulary)
  }

  private val cache = new LRUCache[Formula, Conjunction] (1000)
  
  /**
   * <code>inputFor</code> is the formula to be disproven. The result of the
   * method is a countermodel of <code>inputFor</code>, or <code>FALSE</code>
   * if it was not possible to find one (this implies that <code>inputFor</code>
   * is valid).
   */
  def apply(inputFor : Formula, order : TermOrder) : Conjunction = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ModelSearchProver.AC,
                    inputFor.variables.isEmpty && (order isSortingOf inputFor))
    ////////////////////////////////////////////////////////////////////////////
    cache(inputFor) {
      applyHelp(inputFor, order, GoalSettings.DEFAULT)
    }
  }
   
  def apply(inputFor : Formula, order : TermOrder,
            settings : GoalSettings) : Conjunction = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ModelSearchProver.AC,
                    inputFor.variables.isEmpty && (order isSortingOf inputFor))
    ////////////////////////////////////////////////////////////////////////////
    applyHelp(inputFor, order,
              Param.CONSTRAINT_SIMPLIFIER.set(settings, simplifier))
  }
   
  /**
   * <code>inputFor</code> is the formula to be disproven. The result of the
   * method is a countermodel of <code>inputFor</code>, or <code>None</code>
   * if it was not possible to find one (this implies that <code>inputFor</code>
   * is valid).
   */
  private def applyHelp(inputFor : Formula, order : TermOrder,
                        settings : GoalSettings) : Conjunction = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ModelSearchProver.AC, inputFor.variables.isEmpty)
    ////////////////////////////////////////////////////////////////////////////
    val goal = Goal(inputFor, Set() ++ order.orderedConstants, order, settings)
    try {
      expandProof(goal, Set(), 0)
      Conjunction.FALSE
    } catch {
      case FoundModelExc(facts, constantValues, order, constsToIgnore) => {
        val valuesConj =
          EquationConj(for (c <- FilterIt(order.orderedConstants.elements,
                                          ((c:ConstantTerm) => !(constsToIgnore contains c))))
                       yield LinearCombination(Array((IdealInt.ONE, c),
                                                     (IdealInt.MINUS_ONE, constantValues(c))),
                                               order),
                       order)
        Conjunction.conj(Array(valuesConj, facts), order)
      }
    }
  }

  private def isSatisfiableGoal(tree : ProofTree) : Boolean = tree match {
    case goal : Goal => {
      val c = goal.closingConstraint
      if (c.isTrue) {
        true
      } else if (c.isFalse) {
        false
      } else {
        assert(false) // not supposed to happen
        false
      }
    }
    case _ => false
  }
   
  private def expandProof(tree : ProofTree,
                          // explicitly quantified constants that do not have to
                          // be included in models
                          constsToIgnore : Set[ConstantTerm],
                          depth : Int) : Unit = {
    Timeout.check
      
    if (isSatisfiableGoal(tree)) {
 //     println("backtracking (depth " + depth + ")")
    } else
    tree match { 
      case goal : Goal => {
        if (goal.stepPossible)
          expandProof(goal.step(ptf), constsToIgnore, depth)
        else
          throw FoundModelExc(goal.facts,
                              new IdentitySubst (goal.order),
                              goal.order,
                              constsToIgnore)
      }

      case tree : WitnessTree => {
        // we build a local stack with all witnesses to avoid
        // StackOverflowExceptions
        val witnessStack = new Stack[(Substitution, TermOrder) => Substitution]
                                      
        var subtree : ProofTree = tree
        while (subtree.isInstanceOf[WitnessTree]) {
          val tree = subtree.asInstanceOf[WitnessTree]
          witnessStack push tree.witness
          subtree = tree.subtree
        }

        try {
          expandProof(subtree, constsToIgnore, depth)
          ////////////////////////////////////////
          // some arbitrary operations to avoid tail recursion ...
          // this seems to be a compiler bug
          // TODO: write a bug report
          var x = 0; x = x
        } catch {
          case FoundModelExc(facts, constantValues, order, newConstsToIgnore) => {
            var newCValues = constantValues
            while (!witnessStack.isEmpty)
              newCValues = witnessStack.pop()(newCValues, order)
            
            throw FoundModelExc(facts, newCValues, order, newConstsToIgnore)
          }
        }
      }
      
      case tree : ProofTreeOneChild => {
        ////////////////////////////////////////////////////////////////////////
        Debug.assertInt(ModelSearchProver.AC, tree match {
                          case _ : WeakenTree => false
                          case tree : QuantifiedTree => tree.quan == Quantifier.ALL
                          case _ => true
                        })
        ////////////////////////////////////////////////////////////////////////
        val newConstsToIgnore = tree match {
          case tree : QuantifiedTree => constsToIgnore ++ tree.quantifiedConstants
          case _ => constsToIgnore
        }
        expandProof(tree.subtree, newConstsToIgnore, depth)
      }
     
      case tree : AndTree => {
        expandProof(tree.left, constsToIgnore, depth + 1)
        expandProof(tree.right, constsToIgnore, depth + 1)
      }
      
    }
  }
  
  private case class FoundModelExc(facts : Conjunction,
                                   constantValues : Substitution,
                                   order : TermOrder,
                                   constsToIgnore : Set[ConstantTerm]) extends Exception
}

 
private case class WitnessTree(val subtree : ProofTree,
                               val eliminatedConstant : ConstantTerm,
                               val witness : (Substitution, TermOrder) => Substitution,
                               val vocabulary : Vocabulary)
                   extends { protected val subtreeOrder = vocabulary.order }
                           with ProofTreeOneChild {

  def update(newSubtree : ProofTree, newConstantFreedom : ConstantFreedom) : ProofTree =
    new WitnessTree(subtree, eliminatedConstant, witness,
                    vocabulary updateConstantFreedom newConstantFreedom)

  lazy val closingConstraint : Conjunction =
    subtree.closingConstraint

  lazy val closingConstantFreedom : ConstantFreedom =
    subtree.closingConstantFreedom

  lazy val fixedConstantFreedom : Boolean =
    subtree.fixedConstantFreedom

  lazy val stepMeaningful : Boolean =
    subtree.stepMeaningful
  
  def newConstantFreedomForSubtree(cf : ConstantFreedom) : ConstantFreedom = cf
}